"""
A collection of of reusable and extendable view classes.
"""

from functools import reduce
from django.views.generic import (
    DetailView, CreateView, UpdateView, DeleteView, ListView, TemplateView
)
from django.db.models import Q
from django.http import JsonResponse
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _
from .mixins import OnSuccessMixin
from .forms import FilterForm


class CustomDetailView(DetailView):
    """
    A generic view class providing customized version of the `DetailView` class.
    """
    fields = None
    action_options = {'update': _('Update'), 'delete': _('Delete')}
    template_name = 'vkk/generic/details.html'


class CustomCreateView(OnSuccessMixin, CreateView):
    """
    A generic view class providing customized version of the `CreateView` class.
    """
    on_success = 'create_success'
    template_name = 'vkk/generic/create.html'


class CustomUpdateView(OnSuccessMixin, UpdateView):
    """
    A generic view class providing customized version of the `UpdateView` class.
    """
    on_success = 'update_success'
    template_name = 'vkk/generic/update.html'
    drop_key = 'pk'


class CustomDeleteView(OnSuccessMixin, DeleteView):
    """
    A generic view class providing customized version of the `DeleteView` class.
    """
    on_success = 'delete_success'
    template_name = 'vkk/generic/delete.html'
    drop_key = 'pk'


class CustomSuccessView(OnSuccessMixin, TemplateView):
    """
    A generic view class offering some functionality for successful actions.
    """
    model = None
    on_success = 'default'


class CustomListView(ListView):
    """
    A generic view class providing customized and extended version of the
     `ListView` class.
    """
    keys = None
    fields = None
    paginate_by = 25
    action_options = {'update': _('Update'), 'delete': _('Delete')}
    template_name = 'vkk/generic/list.html'


class CustomFilterView(CustomListView):
    """
    A generic view class extending `CustomListView` with added functionality 
    for filtering objects.
    """
    template_name = 'vkk/generic/filter.html'
    filter_fields = None

    def get_filter_fields(self):
        """
        Returns a list of fields associated with object model used for filtering.
        """
        if self.filter_fields is not None:
            return self.filter_fields
        elif self.fields is not None and self.fields != '__all__':
            return self.fields
        elif self.fields is None:
            raise ImproperlyConfigured(
                'filter_fields or fields not specified.')
        elif self.fields == '__all__':
            raise ImproperlyConfigured(
                "'__all__' not allowed (for security reasons).")

    def get_filter(self):
        """
        Returns a list of arguments for the filter()-method on a manager.
        """
        filter_fields = self.get_filter_fields()
        form = FilterForm(self.request.GET)
        if form.is_valid():
            parameters = form.cleaned_data['query'].split()
        else:
            parameters = []
        if parameters:
            qs = (
                Q(
                    **{field + '__icontains': str(param)}
                ) for field in filter_fields for param in parameters
            )
            query_filter = reduce(lambda q1, q2: q1 | q2, qs)
        else:
            query_filter = ~Q(pk=None)
        return query_filter

    def get_queryset(self):
        """
        Returns a filtered queryset.
        """
        return super().get_queryset().filter(self.get_filter())

    def get(self, request, *args, **kwargs):
        if request.content_type == 'application/json':
            query_values = self.get_queryset().values(
                *self.get_filter_fields()
            )
            suggestions = [' '.join(
                (str(val) for val in obj.values())
            ) for obj in query_values]
            return JsonResponse({'suggestions': suggestions})
        else:
            return super().get(request)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["filter_form"] = FilterForm()
        return context
