// Initialization
const start = document.getElementById("id_start");
const end = document.getElementById("id_end");
const csrftoken = getCookie("csrftoken"); // Required for backend communication

// Get and set nodes for DOM manipulation
const form = document.getElementsByTagName("form")[1];

//// Node for displaying data
const datanode = document.createElement("div");
form.appendChild(datanode);

//// Node for displaying errors
//// This one may already exist and should be overwritten right away.
const errornode = document.getElementsByClassName("errorlist nonfield")[0] || (() => {
    const newErrornode = document.createElement("ul");
    newErrornode.setAttribute("class", "errorlist nonfield");
    newErrornode.setAttribute("style", "display:None;");
    form.prepend(newErrornode);
    return newErrornode;
})();

// Treat page/script loading as first change of input
changeOfInput()

// Add event listeners for change of input
start.addEventListener("change", changeOfInput);
end.addEventListener("change", changeOfInput);

// Functions
//// Checks whether start and end dates are OK and triggers
//// communication with backend and rerender when needed
async function changeOfInput() {
    let newState;
    if (start.value != "", end.value != "") {
        const start_date = new Date(start.value);
        const end_date = new Date(end.value);
        if (start_date <= end_date) {
            newState = await postFormData();
        } else {
            newState = {
                errors: {
                    __all__: ["Start muss vor Ende sein."],
                },
            };
        }
    }
    changeAndRenderState(newState || {});
}

//// Fetches data from backend based on form inputs
async function postFormData() {
    const url = "";
    const data = {
        start: start.value,
        end: end.value,
    };

    const response = await fetch(
        url,
        {
            method: 'POST',
            mode: 'same-origin',
            headers : {
                'Content-Type': 'application/json',
                'X-CSRFToken': csrftoken,
            },
            body: JSON.stringify(data),
        }
    );

    return response.json();
}

//// Rerenders interactive parts
function changeAndRenderState(newState) {
    console.log(newState);
    const {
        errors,
        project,
        department,
        general_costs,
        department_costs,
        project_funded_staff, 
        salary_costs_annotated,
    } = newState;
    renderErrors(errors);

    datanode.replaceChildren();
    if(newState) {
        renderProjectAndDepartment(project, department);
        renderStaffCosts(project_funded_staff, salary_costs_annotated);
        renderDepartmentCosts(department_costs);
        renderGeneralCosts(general_costs);
    }
}

function renderErrors(errors) {
    if(errors && errors.__all__) {
        errornode.removeAttribute("style");
        errornode.replaceChildren();
        const errorArray = errors.__all__;
        errorArray.forEach((element) => {
            const error_item = document.createElement("li");
            error_item.innerText = element;
            errornode.appendChild(error_item);
        })
    } else {
        errornode.replaceChildren();
        errornode.setAttribute("style","display:None;");
    }
}

function renderTableRow(th_text, td_text, parent_node) {
    const row = document.createElement("tr");
    parent_node.appendChild(row);

    const th = document.createElement("th");
    th.innerText = th_text;
    row.appendChild(th);

    const td = document.createElement("td");
    td.innerText = td_text;
    row.appendChild(td);
}

function renderProjectAndDepartment(project, department) {
    if(project && department) {
        const table = document.createElement("table");
        datanode.appendChild(table);

        const caption = document.createElement("caption");
        caption.innerHTML = "Kostenkalkulation für Forschungsprojekte und Dienstleistungen";
        table.appendChild(caption) 

        renderTableRow("Kostenstelle", project.invoice_number, table);
        renderTableRow("Einrichtung", department.name, table);
        renderTableRow("Projekttitel", project.name, table);
        renderTableRow("Vertragspartner", project.contractor, table);
        renderTableRow("Projektlaufzeit", `${project.start} - ${project.end}`, table);
    }
}

function renderTableCell(text, parent) {
    const td = document.createElement("td");
    td.innerText = text;
    parent.appendChild(td);
}

function renderTableHeadCell(text, parent) {
    const th = document.createElement("th");
    th.innerText = text;
    parent.appendChild(th);
}

function renderStaffCosts(project_funded_staff, salary_costs_annotated) {
    if (project_funded_staff || salary_costs_annotated) {
        const table = document.createElement("table");
        datanode.appendChild(table);

        const caption = document.createElement("caption");
        caption.innerHTML = "Personalkosten";
        table.appendChild(caption)
        
        const thead = document.createElement("thead");
        table.appendChild(thead);

        const row = document.createElement("tr");
        thead.appendChild(row);
        renderTableHeadCell("Gehaltsstufe", row);
        renderTableHeadCell("Mitarbeiter-Stunden", row);
        renderTableHeadCell("Stundensatz", row);
        renderTableHeadCell("(indirekte)", row);

        const tbody = document.createElement("tbody");
        table.appendChild(tbody);
        renderProjectFundedStaff(project_funded_staff, tbody);
        renderSalaryCostsAnnotated(salary_costs_annotated, tbody);

        const sumrow = document.createElement("tr");
        tbody.appendChild(sumrow);
        const sumtitle = document.createElement("th");
        sumtitle.setAttribute("colspan", 3);
        sumtitle.innerText = "Summe (indirekte)"
        sumrow.appendChild(sumtitle);
        
        renderTableCell(salary_costs_annotated.costs_sum, sumrow);
    }
}

function renderProjectFundedStaff(project_funded_staff, table) {
    if(project_funded_staff) {
        const head = document.createElement("tr");
        table.appendChild(head);

        const title = document.createElement("th");
        title.innerText = "durch das Projekt finanzierte Mitarbeiter";
        title.setAttribute("colspan", 4);
        head.appendChild(title);

        const hours_by_level = project_funded_staff.hours_by_salary_level;

        for(let key in hours_by_level) {
            const salary_level = hours_by_level[key];
            const row = document.createElement("tr");
            table.appendChild(row);
            renderTableCell(key, row);
            renderTableCell(salary_level["hours"], row)
            renderTableCell(salary_level["brutto_per_hour"], row);
            renderTableCell("", row);
        }
    }
}

function renderSalaryCostsAnnotated(salary_costs_annotated, table) {
    if(salary_costs_annotated) {
        const head = document.createElement("tr");
        table.appendChild(head);

        const title = document.createElement("th");
        title.innerText = "am Projekt beteiligte Mitarbeiter (haushaltsfinanziert)";
        title.setAttribute("colspan", 4);
        head.appendChild(title);

        const salary_levels = salary_costs_annotated.salary_levels;

        for(let key in salary_levels) {
            const salary_level = salary_levels[key];
            const row = document.createElement("tr");
            table.appendChild(row);
            renderTableCell(key, row);
            renderTableCell(salary_level["hours"], row)
            renderTableCell(salary_level["brutto_per_hour"], row);
            renderTableCell(salary_level["costs"], row);
        }
    }
}

function renderDepartmentCosts(department_costs) {
    if(department_costs) {
        const table = document.createElement("table");
        datanode.appendChild(table);

        const caption = document.createElement("caption");
        caption.innerHTML = "Sachkosten (Tab.-Blatt \"Sachkosten\" verwenden)";
        table.appendChild(caption)

        const thead = document.createElement("thead");
        table.appendChild(thead);

        const head_row = document.createElement("tr");
        thead.appendChild(head_row)
        renderTableHeadCell("pro h / VZÄ", head_row)
        renderTableHeadCell("indirekt", head_row)

        const tbody = document.createElement("tbody")
        table.appendChild(tbody);
        const body_row = document.createElement("tr");
        tbody.append(body_row);
        renderTableCell(department_costs.equivalents_per_hour, body_row);
        renderTableCell(department_costs.salary_costs, body_row);
    }
}

function renderGeneralCosts(general_costs) {
    if(general_costs) {
        const table = document.createElement("table");
        datanode.appendChild(table);

        const caption = document.createElement("caption");
        caption.innerHTML = "Sach- und Sonstige Gemeinkosten Med. Fakultät";
        table.appendChild(caption)

        const thead = document.createElement("thead");
        table.appendChild(thead);

        const head_row = document.createElement("tr");
        thead.appendChild(head_row)
        renderTableHeadCell("Summe GK (SachGK + Sonst. GK pro VZÄ/h)", head_row)
        renderTableHeadCell("Sach- und Sonstige Gemeinkosten gesamt:", head_row)

        const tbody = document.createElement("tbody")
        table.appendChild(tbody);
        const body_row = document.createElement("tr");
        tbody.append(body_row);
        renderTableCell(general_costs.costs, body_row);
        renderTableCell(general_costs.total, body_row);
    }
}


//// Helper function for retrieving a cookie value
//// This is needed for getting the csrf token
function getCookie(name) {
    let cookieValue = null;
    if (document.cookie && document.cookie !== '') {
        const cookies = document.cookie.split(';');
        for (let i = 0; i < cookies.length; i++) {
            const cookie = cookies[i].trim();
            // Does this cookie string begin with the name we want?
            if (cookie.substring(0, name.length + 1) === (name + '=')) {
                cookieValue = decodeURIComponent(cookie.substring(name.length + 1));
                break;
            }
        }
    }
    return cookieValue;
}