"""
A collection of tags and filters used throughout the project.
"""
from django import template
from django.urls import reverse
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from vkk.system.models import Announcement
from .navigation_tree import dirs

register = template.Library()


@register.inclusion_tag('vkk/customtags/announcements.html')
def announcements():
    """
    A tag for adding a div-element containing the currently active system announcements.
    """
    return {'messages': Announcement.objects.filter(is_active=True)}


@register.inclusion_tag('vkk/customtags/form.html')
def div_form(form):
    """
    A tag for rendering a form as a collection of div elements.
     This was introduced prior to Django 4.1.
    """
    return {'form': form}


@register.filter
def verbose_name(obj):
    """
    A filter for getting an objects verbose name.
    """
    return obj._meta.verbose_name


@register.filter
def verbose_name_plural(obj):
    """
    A filter for getting an objects verbose name in a plural form.
    """
    return obj._meta.verbose_name_plural


@register.filter
def handle_value(obj):
    """
    A filter for displaying some python types in a sensible manner.
    """
    if obj is None:
        return _('-')
    elif isinstance(obj, bool):
        return _('Yes') if obj else _('No')
    else:
        return obj


@register.simple_tag
def model_name(model):
    """
    A filter for getting a models verbose name.
    """
    return model._meta.verbose_name


@register.inclusion_tag('vkk/customtags/nav.html', takes_context=True)
def navigation(context):
    """
    A tag for generating and adding a navigation.
    """
    match = context['view'].request.resolver_match
    links = []
    logout = False
    namespace_joined = None
    for namespace in match.namespaces:
        if namespace_joined is None:
            namespace_joined = namespace
        else:
            namespace_joined += ':' + namespace
        entries = dirs.get(namespace_joined)
        if entries is not None:
            for entry in entries:
                kwargs = {}
                for kwarg in entry[2]:
                    kwargs[kwarg] = match.kwargs[kwarg]
                links += [(
                    reverse(
                        namespace_joined + ':' + entry[0],
                        kwargs=kwargs
                    ),
                    entry[1]
                )]

    if context['view'].request.user.is_authenticated:
        links += [
            (reverse('vkk:users:user_details_change'), _('Profile')),
        ]
        logout = True

    context['links'] = links
    context['logout'] = logout
    return context


def _get_option_url(request, option, kwarg):
    match = request.resolver_match
    kwargs = {**match.kwargs, **kwarg}
    namespace = ':'.join(match.namespaces + [option])
    return reverse(namespace, kwargs=kwargs)


@register.inclusion_tag('vkk/customtags/list_table.html')
def list_table(query_set, model, fields, options=None, request=None, key='pk'):
    """
    A tag for generating a HTML table based on a given queryset.
    """

    # Gathers desired fields of the model
    _fields = [model._meta.get_field(field) for field in fields]

    # Generates table head
    thead = [_(field.verbose_name) for field in _fields]
    if options is not None:
        thead += ['' for option in options]

    # Generates table body
    tbody = []
    for instance in query_set:
        row = []
        for field in _fields:
            cell = getattr(instance, field.name)
            row.append(handle_value(cell))
        if options is not None:
            kwarg = {key: getattr(instance, key)}
            for option, value in options.items():
                url = _get_option_url(request, option, kwarg)
                tag = mark_safe('<a href="%s">%s</a>' % (url, escape(value)))
                row.append(tag)
        tbody.append(row)

    return {'head': thead, 'tbody': tbody}


@register.inclusion_tag('vkk/customtags/details_table.html')
def details_table(model_instance, fields):
    """
    A tag for generating a HTML table based on a given model instance.
    """
    _fields = [model_instance._meta.get_field(field) for field in fields]

    table_content = []
    for field in _fields:
        th = field.verbose_name
        td = getattr(model_instance, field.name)
        table_content.append([th, handle_value(td)])

    return {'table_content': table_content}
