"""
A collection of forms for user management.
"""
from django.contrib.auth.forms import SetPasswordForm
from django.forms import ModelForm, CharField, PasswordInput
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from .models import User


class ConfirmNewUserForm(SetPasswordForm, ModelForm):
    """
    A form for confirming a user registration.
    """
    class Meta:
        """
        Meta information for this form class.
        """
        model = User
        fields = ['first_name', 'last_name']

    def __init__(self, *args, **kwargs):
        """
        Constructs a `ConfirmNewUserForm` object.
        """
        kwargs.update({'user': kwargs['instance']})
        super().__init__(*args, **kwargs)

    def save(self, commit=True):
        """
        Saves and returns a `User` instance. 
        """
        instance = super().save(commit)
        instance.is_active = True
        if commit:
            instance.save()
        return instance


class UserDetailsChangeForm(ModelForm):
    """
    A form for updating some user details.
    """
    class Meta:
        """
        Meta information for this form class.
        """
        model = User
        fields = ['first_name', 'last_name']

    error_messages = {
        'password_incorrect': _("Your password was entered incorrectly. Please enter it again."),
    }
    password = CharField(
        label=_("Password"),
        strip=False,
        widget=PasswordInput(
            attrs={'autocomplete': 'current-password', 'autofocus': True}),
        help_text=_('Please enter your password to confirm any changes made.')
    )

    def clean_password(self):
        """
        Validates and returns a string entered in the `password` field.
        """
        password = self.cleaned_data["password"]
        if not self.instance.check_password(password):
            raise ValidationError(
                self.error_messages['password_incorrect'],
                code='password_incorrect',
            )
        return password
