"""
A collection of views for managing Users.
"""
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.views.generic import FormView, TemplateView
from django.core.signing import Signer, BadSignature
from django.contrib.auth.mixins import LoginRequiredMixin
from .models import SALT, NewUser
from .forms import ConfirmNewUserForm, UserDetailsChangeForm


class RegistrationConfirmView(FormView):
    """
    A view class for confirming the registration of a user.
    """
    template_name = 'vkk/users/registration_confirm.html'
    form_class = ConfirmNewUserForm
    success_url = reverse_lazy('vkk:users:registration_confirm_done')

    def setup(self, request, *args, **kwargs):
        """
        Initialize attributes shared by all view methods.
        """
        super().setup(request, *args, **kwargs)
        try:
            token = Signer(sep='/', salt=SALT).unsign(
                '/'.join([kwargs['token'], kwargs['signature']])
            )
            self.new_user = NewUser.objects.get(token=token)
            self.is_link_valid = True
        except (BadSignature, NewUser.DoesNotExist):
            self.is_link_valid = False

    def dispatch(self, request, *args, **kwargs):
        """
        Dispatches a request according to its method.
        """
        if not self.is_link_valid:
            return redirect('vkk:users:invalid_registration')
        else:
            return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        """
        Returns the key word arguments for instantiating the associated
        form class.
        """
        kwargs = super().get_form_kwargs()
        kwargs.update({'instance': self.new_user.user})
        return kwargs

    def form_valid(self, form):
        """
        Tries to save and return the model instance associated with
        the form instance of this class's instance.
        As a result, a new `User` instance is created and a `NewUser` is deleted.
        """
        form.save()
        self.new_user.delete()
        return super().form_valid(form)


class UserDetailsChangeView(LoginRequiredMixin, FormView):
    """
    A view for updating a users associated data and credentials.
    """
    template_name = 'vkk/users/user_details_change.html'
    form_class = UserDetailsChangeForm

    def get_form_kwargs(self):
        """
        Returns the key word arguments for instantiating the associated
        form class.
        """
        kwargs = super().get_form_kwargs()
        kwargs.update({'instance': self.request.user})
        return kwargs

    def form_valid(self, form):
        """
        Tries to save and return the model instance associated with
        the form instance of this class's instance.
        """
        form.save()
        return super().form_valid(form)


class UserDetailsChangeSuccessView(LoginRequiredMixin, TemplateView):
    """
    This view is used for confirming a successful update of a users data or credentials.
    """
    template_name = 'vkk/users/user_details_change_done.html'
