"""
This submodule contains class based views.
"""

from django.shortcuts import get_object_or_404
from vkk.workhours.accounting.views import AccountingDetailView
from vkk.workhours.models import Period, Project, Department


class AccountingPeriodDetailView(AccountingDetailView):
    """
    This class based `View` provides an overview over a `Period` and whether
    all projects for a department have been closed for said period.
    """
    model = Period
    fields = ['start', 'end', 'dead_line', 'dead_line_final']
    template_name = "vkk/workhours/accounting/period/details.html"

    def get_context_data(self, **kwargs):
        """
        Returns a query set of `Departments` and whether all projects for the given
        department and period have been closed.
        """
        context = super().get_context_data(**kwargs)
        departments = Department.objects.raw(" \
            WITH selected_period AS ( \
                SELECT \
                    vkk_period.id AS period_id, \
                    vkk_period.start AS period_start, \
                    vkk_period.end AS period_end \
                FROM vkk_period \
                WHERE vkk_period.id = %s \
            ), \
            filtered_projects AS ( \
                SELECT \
                    vkk_project.id AS project_id, \
                    vkk_project.department_id, \
                    selected_period.period_id \
                FROM vkk_project, selected_period \
                WHERE  \
                    selected_period.period_end >= vkk_project.start  \
                    AND selected_period.period_start <= vkk_project.end \
            ), \
            closures AS ( \
                SELECT \
                    vkk_periodclosure.project_assignment_id, \
                    vkk_periodclosure.is_closed_manager \
                FROM  \
                    vkk_periodclosure JOIN selected_period \
                    ON vkk_periodclosure.period_id = selected_period.period_id \
            ), \
            not_closed_assignments AS ( \
                SELECT \
                    vkk_projectassignment.id AS assignment_id, \
                    vkk_projectassignment.project_id \
                FROM \
                    vkk_projectassignment LEFT JOIN closures \
                    ON vkk_projectassignment.id = closures.project_assignment_id \
                WHERE \
                    closures.is_closed_manager IS NULL \
                    OR closures.is_closed_manager = FALSE \
            ), \
            not_closed_projects AS ( \
                SELECT \
                    department_id, \
                    COUNT(project_id) AS not_closed \
                FROM \
                    filtered_projects p \
                WHERE  \
                    EXISTS ( \
                        SELECT assignment_id \
                        FROM not_closed_assignments c \
                        WHERE  \
                            c.project_id = p.project_id \
                    ) \
                GROUP BY department_id \
            ) \
            SELECT \
                id, \
                name, \
                accounting_entry, \
                invoice_number, \
                COALESCE(not_closed, 0) AS projects_not_closed \
                FROM \
                    vkk_department FULL JOIN not_closed_projects \
                    ON vkk_department.id = not_closed_projects.department_id \
                ORDER BY vkk_department.name \
            ;", [self.kwargs['pk']])

        context["departments"] = departments
        return context


class AccountingPeriodDetailDepartmentView(AccountingDetailView):
    """
    A class based `View` offering an overview of all projects belonging to a 
    given department over a given period and whether they have been closed.
    """
    model = Period
    fields = ['start', 'end', 'dead_line', 'dead_line_final']
    template_name = "vkk/workhours/accounting/period/projects.html"

    def get_context_data(self, **kwargs):
        """
        Returns a query set of `Projects` over a period and beloginging to a given
        department with the added information whether they have been closed or not.
        """
        context = super().get_context_data(**kwargs)
        department = get_object_or_404(
            Department, pk=self.kwargs["department_pk"])
        projects_open = Project.objects.raw(" \
            WITH selected_period AS ( \
                SELECT \
                    vkk_period.id AS period_id, \
                    vkk_period.start AS period_start, \
                    vkk_period.end AS period_end \
                FROM vkk_period \
                WHERE vkk_period.id = %s \
            ), \
            filtered_projects AS ( \
                SELECT \
                    selected_period.period_id, \
                    vkk_project.id, \
                    vkk_project.invoice_number, \
                    vkk_project.name, \
                    vkk_project.contractor, \
                    vkk_project.start, \
                    vkk_project.end, \
                    vkk_project.department_id \
                FROM vkk_project, selected_period \
                WHERE  \
                    selected_period.period_end >= vkk_project.start  \
                    AND selected_period.period_start <= vkk_project.end \
                    AND vkk_project.department_id = %s \
            ), \
            closures AS ( \
                SELECT \
                    vkk_periodclosure.project_assignment_id, \
                    vkk_periodclosure.is_closed_manager \
                FROM  \
                    vkk_periodclosure JOIN selected_period \
                    ON vkk_periodclosure.period_id = selected_period.period_id \
            ), \
            not_closed_assignments AS ( \
                SELECT \
                    vkk_projectassignment.id AS assignment_id, \
                    vkk_projectassignment.project_id \
                FROM \
                    vkk_projectassignment LEFT JOIN closures \
                    ON vkk_projectassignment.id = closures.project_assignment_id \
                WHERE \
                    closures.is_closed_manager IS NULL \
                    OR closures.is_closed_manager = FALSE \
            ) \
            SELECT \
                p.id, \
                p.invoice_number, \
                p.name, \
                p.contractor, \
                p.start, \
                p.end, \
                p.department_id \
            FROM \
                filtered_projects p \
            WHERE  \
                EXISTS ( \
                    SELECT assignment_id \
                    FROM not_closed_assignments c \
                    WHERE  \
                        c.project_id = p.id \
                ) \
            ORDER BY p.invoice_number;",
            [self.kwargs['pk'], self.kwargs["department_pk"]]
        )

        projects_closed = Project.objects.raw(" \
            WITH selected_period AS ( \
                SELECT \
                    vkk_period.id AS period_id, \
                    vkk_period.start AS period_start, \
                    vkk_period.end AS period_end \
                FROM vkk_period \
                WHERE vkk_period.id = %s \
            ), \
            filtered_projects AS ( \
                SELECT \
                    selected_period.period_id, \
                    vkk_project.id, \
                    vkk_project.invoice_number, \
                    vkk_project.name, \
                    vkk_project.contractor, \
                    vkk_project.start, \
                    vkk_project.end, \
                    vkk_project.department_id, \
                    selected_period.period_start, \
                    selected_period.period_end \
                FROM vkk_project, selected_period \
                WHERE  \
                    selected_period.period_end >= vkk_project.start  \
                    AND selected_period.period_start <= vkk_project.end \
                    AND vkk_project.department_id = %s \
            ), \
            closures AS ( \
                SELECT \
                    vkk_periodclosure.project_assignment_id, \
                    vkk_periodclosure.is_closed_manager \
                FROM  \
                    vkk_periodclosure JOIN selected_period \
                    ON vkk_periodclosure.period_id = selected_period.period_id \
            ), \
            not_closed_assignments AS ( \
                SELECT \
                    vkk_projectassignment.id AS assignment_id, \
                    vkk_projectassignment.project_id \
                FROM \
                    vkk_projectassignment LEFT JOIN closures \
                    ON vkk_projectassignment.id = closures.project_assignment_id \
                WHERE \
                    closures.is_closed_manager IS NULL \
                    OR closures.is_closed_manager = FALSE \
            ) \
            SELECT \
                p.id, \
                p.invoice_number, \
                p.name, \
                p.contractor, \
                p.start, \
                p.end, \
                p.department_id, \
                CASE  \
                    WHEN EXISTS ( \
                        SELECT 1 \
                            FROM vkk_receipt \
                            WHERE \
                                p.id = vkk_receipt.project_id \
                                AND p.period_start = vkk_receipt.start \
                                AND p.period_end = vkk_receipt.end \
                    ) THEN TRUE \
                    ELSE FALSE \
                END AS exists_receipt \
            FROM \
                filtered_projects p \
            WHERE  \
                NOT EXISTS ( \
                    SELECT assignment_id \
                    FROM not_closed_assignments c \
                    WHERE  \
                        c.project_id = p.id \
                ) \
            ORDER BY p.invoice_number;",
            [self.kwargs['pk'], self.kwargs["department_pk"]]
        )

        context["department"] = department
        context["projects_open"] = projects_open
        context["projects_closed"] = projects_closed
        return context

class AccountingPeriodDetailNoDepartmentView(AccountingDetailView):
    """
    A class based `View` providing an overview of projects which do not belong to any
    department for the given period.
    """
    model = Period
    fields = ['start', 'end', 'dead_line', 'dead_line_final']
    template_name = "vkk/workhours/accounting/period/projects.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        projects_open = Project.objects.raw(" \
            WITH selected_period AS ( \
                SELECT \
                    vkk_period.id AS period_id, \
                    vkk_period.start AS period_start, \
                    vkk_period.end AS period_end \
                FROM vkk_period \
                WHERE vkk_period.id = %s \
            ), \
            filtered_projects AS ( \
                SELECT \
                    selected_period.period_id, \
                    vkk_project.id, \
                    vkk_project.invoice_number, \
                    vkk_project.name, \
                    vkk_project.contractor, \
                    vkk_project.start, \
                    vkk_project.end, \
                    vkk_project.department_id \
                FROM vkk_project, selected_period \
                WHERE  \
                    selected_period.period_end >= vkk_project.start  \
                    AND selected_period.period_start <= vkk_project.end \
                    AND vkk_project.department_id IS NULL \
            ), \
            closures AS ( \
                SELECT \
                    vkk_periodclosure.project_assignment_id, \
                    vkk_periodclosure.is_closed_manager \
                FROM  \
                    vkk_periodclosure JOIN selected_period \
                    ON vkk_periodclosure.period_id = selected_period.period_id \
            ), \
            not_closed_assignments AS ( \
                SELECT \
                    vkk_projectassignment.id AS assignment_id, \
                    vkk_projectassignment.project_id \
                FROM \
                    vkk_projectassignment LEFT JOIN closures \
                    ON vkk_projectassignment.id = closures.project_assignment_id \
                WHERE \
                    closures.is_closed_manager IS NULL \
                    OR closures.is_closed_manager = FALSE \
            ) \
            SELECT \
                p.id, \
                p.invoice_number, \
                p.name, \
                p.contractor, \
                p.start, \
                p.end, \
                p.department_id \
            FROM \
                filtered_projects p \
            WHERE  \
                EXISTS ( \
                    SELECT assignment_id \
                    FROM not_closed_assignments c \
                    WHERE  \
                        c.project_id = p.id \
                ) \
            ORDER BY p.invoice_number;",
            [self.kwargs['pk']]
        )

        projects_closed = Project.objects.raw(" \
            WITH selected_period AS ( \
                SELECT \
                    vkk_period.id AS period_id, \
                    vkk_period.start AS period_start, \
                    vkk_period.end AS period_end \
                FROM vkk_period \
                WHERE vkk_period.id = %s \
            ), \
            filtered_projects AS ( \
                SELECT \
                    selected_period.period_id, \
                    vkk_project.id, \
                    vkk_project.invoice_number, \
                    vkk_project.name, \
                    vkk_project.contractor, \
                    vkk_project.start, \
                    vkk_project.end, \
                    vkk_project.department_id, \
                    selected_period.period_start, \
                    selected_period.period_end \
                FROM vkk_project, selected_period \
                WHERE  \
                    selected_period.period_end >= vkk_project.start  \
                    AND selected_period.period_start <= vkk_project.end \
                    AND vkk_project.department_id IS NULL \
            ), \
            closures AS ( \
                SELECT \
                    vkk_periodclosure.project_assignment_id, \
                    vkk_periodclosure.is_closed_manager \
                FROM  \
                    vkk_periodclosure JOIN selected_period \
                    ON vkk_periodclosure.period_id = selected_period.period_id \
            ), \
            not_closed_assignments AS ( \
                SELECT \
                    vkk_projectassignment.id AS assignment_id, \
                    vkk_projectassignment.project_id \
                FROM \
                    vkk_projectassignment LEFT JOIN closures \
                    ON vkk_projectassignment.id = closures.project_assignment_id \
                WHERE \
                    closures.is_closed_manager IS NULL \
                    OR closures.is_closed_manager = FALSE \
            ) \
            SELECT \
                p.id, \
                p.invoice_number, \
                p.name, \
                p.contractor, \
                p.start, \
                p.end, \
                p.department_id, \
                CASE  \
                    WHEN EXISTS ( \
                        SELECT 1 \
                            FROM vkk_receipt \
                            WHERE \
                                p.id = vkk_receipt.project_id \
                                AND p.period_start = vkk_receipt.start \
                                AND p.period_end = vkk_receipt.end \
                    ) THEN TRUE \
                    ELSE FALSE \
                END AS exists_receipt \
            FROM \
                filtered_projects p \
            WHERE  \
                NOT EXISTS ( \
                    SELECT assignment_id \
                    FROM not_closed_assignments c \
                    WHERE  \
                        c.project_id = p.id \
                ) \
            ORDER BY p.invoice_number;",
            [self.kwargs['pk']]
        )

        context["department"] = None
        context["projects_open"] = projects_open
        context["projects_closed"] = projects_closed
        return context