"""
This submodule contains class based views.
"""

import json
import csv
import datetime
from django.utils.translation import gettext_lazy as _
from django.http import JsonResponse, HttpResponse
from django.views.generic import FormView
from django.shortcuts import get_object_or_404
from vkk.workhours.accounting.mixins import AccountantRequiredMixin
from vkk.workhours.accounting.views import (
    AccountingListView, AccountingCreateView
)
from vkk.workhours.models import Receipt, Project
from .forms import ReceiptForm, ReceiptTemplateSelectForm


class ReceiptListView(AccountingListView):
    """
    A class based `ListView` providing utilities for listing `Receipt`s.
    """
    model = Receipt
    fields = ['create_time', 'start', 'end']
    ordering = ['create_time']
    action_options = {
        'details': _('Details'),
        'delete': _('Delete'),
        'receipt_download': _('Download'),
    }

    def get_queryset(self):
        """
        Returns a query set of `Receipt`s for a given project.
        """
        return super().get_queryset().filter(
            project__invoice_number=self.kwargs['invoice_number'],
        )


class ReceiptCreateView(AccountingCreateView):
    """
    A class based `CreateView` providing utilities for creating `Receipt`s.
    """
    model = Receipt
    form_class = ReceiptForm

    def post(self, request, *args, **kwargs):
        """
        Handler for POST requests.
        """
        if request.content_type == 'application/json':
            return self.post_json(request, *args, **kwargs)
        else:
            return super().post(request, *args, **kwargs)

    def post_json(self, request, *args, **kwargs):
        """
        Handler for POST requests with content type `'application/json'`.
        """
        kwargs = self.get_form_kwargs()
        kwargs.update({
            "data": json.loads(request.body),
        })
        form = self.get_form_class()(**kwargs)

        if form.is_valid():
            return JsonResponse(form.to_data_dict())
        else:
            return JsonResponse(
                {'errors': form.errors}
            )

    def get_form_kwargs(self):
        """
        Returns a dictionary of keywords for instanciating the associated form.
        """
        kwargs = super().get_form_kwargs()
        project = Project.objects.get(
            invoice_number=self.kwargs['invoice_number']
        )
        kwargs.update({
            "project": project,
        })
        return kwargs


class ReceiptTemplateSelectionView(FormView, AccountantRequiredMixin):
    """
    A class based `View` providing utilities for selecting a template for a receipt and export
    in shape of a `.csv` file.
    """
    form_class = ReceiptTemplateSelectForm
    template_name = "vkk/workhours/accounting/projects/project/export/receipt_select.html"

    def form_valid(self, form):
        """
        Returns a CSV-Response based on the given receipt and receipt template from the form.
        """
        receipt_template = form.cleaned_data.get("receipt_template")
        receipt = get_object_or_404(Receipt, pk=self.kwargs["pk"])
        return self.csv_response(receipt, receipt_template)

    def csv_response(self, receipt, receipt_template):
        """
        Returns a CSV-Response based on the given receipt and receipt template.
        """
        response = HttpResponse(
            content_type='text/csv',
            headers={
                'Content-Disposition': f'attachment; filename="{ receipt.id }.csv"'},
        )

        # init writer for creating csv
        writer = csv.writer(response, delimiter=';')

        # retrieving data for creating the receipt in csv format
        template_data = receipt_template.data
        data = json.loads(receipt.data)

        # retrieving keys and writing header row
        keys = [list(column.keys())[0] for column in template_data]
        writer.writerow([str(key) for key in keys])

        # getting columns and reorganize them to rows
        table_template = zip(*(column.get(key)
                             for column, key in zip(template_data, keys)))

        # here be conversion
        table = [
            [
                _data_cell_calculation(cell, receipt, data) for cell in row
            ] for row in table_template
        ]

        # writing data to csv
        writer.writerows(table)

        return response


def _data_cell_calculation(template, receipt, data):
    """This functions retrieves the required data point according to the template"""
    method = template.get("method")
    if method is None:
        return ""
    elif method == "FIXED":
        return template.get("value")
    elif method == "GENERATED":
        value = template.get("value")
        args = value.split(" ")
        out = []
        for arg in args:
            lookup = None
            parts = arg.split(".")
            if parts[0] == "data":
                lookup = data
            elif hasattr(receipt, parts[0]):
                lookup = getattr(receipt, parts[0])
            if lookup is not None:
                for part in parts[1:]:
                    if isinstance(lookup, dict):
                        lookup = lookup.get(part)
                    elif hasattr(lookup, part):
                        lookup = getattr(lookup, part)
                    if lookup is None:
                        break
                if lookup is not None:
                    out.append(str(_format(lookup)))
            else:
                out.append(str(arg))
        return " ".join(out)
    else:
        return ""


def _format(value):
    if isinstance(value, datetime.date):
        return value.strftime("%d.%m.%Y")
    elif isinstance(value, str):
        return 0 if value == "0.00" else value
    elif isinstance(value, bool):
        return _("Ja") if value else _("Nein")
    else:
        return value
