"""
This submodule contains the routing configuration.
"""

from django.urls import include, path
from django.utils.translation import gettext_lazy as _
from django.forms import modelform_factory
from vkk.workhours.models import ProjectManager, ProjectAssignment, Project, WorkHours
from vkk.generic.forms import CustomDateInput
from vkk.workhours.accounting.views import *
from .views import *

key = {'slug_field': 'invoice_number', 'slug_url_kwarg': 'invoice_number'}

app_name = 'project'
urlpatterns = [
    path(
        _('details/'),
        AccountingProjectOverView.as_view(),
        name='default'
    ),
    path(
        _('update/'),
        AccountingProjectUpdateView.as_view(),
        name='update'
    ),
    path(
        _('update/success/'),
        AccountingSuccessView.as_view(
            model=Project,
            template_name='vkk/generic/update_success.html',
        ),
        name='update_success'
    ),
    # manager
    path(
        _('manager/<int:pk>/details/'),
        AccountingDetailView.as_view(
            model=ProjectManager,
            template_name='vkk/workhours/accounting/projects/project/manager_details.html',
        ),
        name='manager_details'
    ),
    path(
        _('manager/<int:pk>/update/'),
        AccountingUpdateView.as_view(
            model=ProjectManager,
            form_class=modelform_factory(
                model=ProjectManager,
                fields=['start', 'end', 'is_proxy'],
                widgets={'start': CustomDateInput, 'end': CustomDateInput}
            ),
            on_success='manager_update_success',
        ),
        name='manager_update'
    ),
    path(
        _('manager/update/success/'),
        AccountingSuccessView.as_view(
            model=ProjectManager,
            template_name='vkk/generic/update_success.html',
        ),
        name='manager_update_success'
    ),
    path(
        _('manager_candidates/'),
        AccountingProjectUserFilterView.as_view(
            to_exclude='projectmanager__project__invoice_number',
            action_options={'add_manager': _('Add')},
            ordering=['last_name', 'first_name'],
        ),
        name='manager_candidates'
    ),
    path(
        _('manager_candidates/<int:pk>/add_new/'),
        AccountingManagerCreateView.as_view(),
        name='add_manager'
    ),
    path(
        _('manager_candidates/add_new/success/'),
        AccountingSuccessView.as_view(
            model=ProjectManager,
            template_name='vkk/generic/create_success.html',
            on_success='manager_candidates'
        ),
        name='add_manager_success'
    ),
    path(
        _('manager/<int:pk>/delete'),
        AccountingManagerDeleteView.as_view(),
        name='delete_manager'
    ),
    path(
        _('manager/delete/success/'),
        AccountingSuccessView.as_view(
            model=ProjectAssignment,
            template_name='vkk/generic/delete_success.html',
            on_success='default'
        ),
        name='delete_manager_success'
    ),
    # contributor
    path(
        _('contributor/<int:pk>/details/'),
        AccountingDetailView.as_view(
            model=ProjectAssignment,
            template_name='vkk/workhours/accounting/projects/project/contributor_details.html',
        ),
        name='contributor_details'
    ),
    path(
        _('contributor/<int:pk>/update/'),
        AccountingUpdateView.as_view(
            model=ProjectAssignment,
            fields=['salary_level'],
            on_success='contributor_update_success',
        ),
        name='contributor_update'
    ),
    path(
        _('contributor/update/success/'),
        AccountingSuccessView.as_view(
            model=ProjectManager,
            template_name='vkk/generic/update_success.html',
        ),
        name='contributor_update_success'
    ),
    path(
        _('contributor_candidates/'),
        AccountingProjectUserFilterView.as_view(
            to_exclude='projectassignment__project__invoice_number',
            action_options={'add_contributor': _('Add')},
            ordering=['last_name', 'first_name'],
        ),
        name='contributor_candidates'
    ),
    path(
        _('contributor_candidates/<int:pk>/add_new/'),
        AccountingContributorCreateView.as_view(),
        name='add_contributor'
    ),
    path(
        _('contributor_candidates/add_new/success/'),
        AccountingSuccessView.as_view(
            model=ProjectAssignment,
            template_name='vkk/generic/create_success.html',
            on_success='contributor_candidates'
        ),
        name='add_contributor_success'
    ),
    path(
        _('contributor/<int:pk>/delete'),
        AccountingContributorDeleteView.as_view(),
        name='delete_contributor'
    ),
    path(
        _('contributor/delete/success/'),
        AccountingSuccessView.as_view(
            model=ProjectAssignment,
            template_name='vkk/generic/delete_success.html',
            on_success='default'
        ),
        name='delete_contributor_success'
    ),
    # workhour sheet
    path(
        _('workhours_sheet/'),
        AccountingWorkhourSheetSelectionView.as_view(),
        name='workhours_sheet_selection'
    ),
    path(
        _('workhours_sheet/<int:period_pk>/'),
        AccountingWorkhourSheetView.as_view(),
        name='workhours_sheet'
    ),
    path(
        _('workhours_sheet/<int:period_pk>/success/'),
        AccountingSuccessView.as_view(
            template_name='vkk/workhours/workhours_sheet_success.html',
            model=WorkHours,
            on_success='workhours_sheet',
        ),
        name='workhours_sheet_success'
    ),
    path(
        _('workhours_sheet/<int:period_pk>/closure/'),
        AccountingPeriodClosureView.as_view(),
        name='period_closure'
    ),
    path(
        _('workhours_sheet/<int:period_pk>/closure/success/'),
        AccountingPeriodClosureSuccessView.as_view(),
        name='period_closure_success'
    ),
    # export
    path(_('export/'), include('vkk.workhours.accounting.projects.project.export.urls')),
]
