"""
This submodule contains class based views.
"""

from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError
from django.forms import modelform_factory, Form
from django.urls import reverse
from django.shortcuts import get_object_or_404
from django.views.generic import RedirectView, FormView, TemplateView
from vkk.generic.forms import CustomDateInput
from vkk.generic.mixins import OnSuccessMixin
from vkk.generic.views import CustomSuccessView
from vkk.users.models import User
from vkk.workhours.models import ProjectManager, ProjectAssignment, Project, Period, PeriodClosure, WorkHours, WorkHoursCorrection
from vkk.workhours.mixins import PeriodSelectorMixin
from vkk.workhours.forms import WorkhourSheetForm, PeriodSelectForm
from vkk.workhours.accounting.mixins import AccountantRequiredMixin
from vkk.workhours.accounting.views import AccountingDetailView, AccountingFilterView, AccountingCreateView, AccountingUpdateView, AccountingDeleteView
from vkk.workhours.accounting.projects.project.forms import ContributorDeleteForm


class AccountingProjectOverView(AccountingDetailView):
    """
    A class based `View` providing  an overview over the given project.
    """
    model = Project
    slug_field = 'invoice_number'
    slug_url_kwarg = 'invoice_number'
    template_name = 'vkk/workhours/accounting/projects/project/details.html'
    fields = ['name', 'invoice_number',
              'department', 'contractor', 'start', 'end']

    def get_context_data(self, **kwargs):
        """
        Returns additional context data as an dictionary. This is used for rendering the page.
        """
        context = super().get_context_data(**kwargs)
        context["managers"] = ProjectManager.objects.filter(
            project__invoice_number=self.kwargs['invoice_number']
        ).prefetch_related('manager').order_by('manager__last_name', 'manager__first_name')
        context["contributors"] = ProjectAssignment.objects.filter(
            project__invoice_number=self.kwargs['invoice_number']
        ).prefetch_related('contributor').order_by('contributor__last_name', 'contributor__first_name')
        return context


class AccountingProjectUserFilterView(AccountingFilterView):
    """
    A class based `View` providing utilities for selecting a `User` which
    is to be assigned to the given project.
    """
    model = User
    fields = ['last_name', 'first_name', 'email']
    action_options = None
    to_exclude = None

    def get(self, request, *args, invoice_number=None, **kwargs):
        """
        Handler for GET requests.
        """
        return super().get(request, *args, **kwargs)

    def get_queryset(self):
        """
        Returns a query set of `Users` which are not assigned to the given Project.
        """
        kwarg = {self.to_exclude: self.kwargs['invoice_number']}
        return super().get_queryset().exclude(**kwarg)


class AccountingProjectUpdateView(AccountingUpdateView):
    """
    A class based `UpdateView` for updating the given project.
    """
    model = Project
    form_class = modelform_factory(
        model=Project,
        fields=['invoice_number', 'name', 'department',
                'contractor', 'start', 'end'],
        widgets={'start': CustomDateInput, 'end': CustomDateInput}
    )
    drop_key = None
    slug_field = 'invoice_number'
    slug_url_kwarg = 'invoice_number'

    def form_valid(self, form):
        """
        Saves any changes to the associated `Project`.
        """
        self.kwarg_override = {'invoice_number': form.instance.invoice_number}
        return super().form_valid(form)


class AccountingManagerCreateView(AccountingCreateView):
    """
    A class based `CreateView` providing utilities for creating new `ProjectManager`s.
    """
    model = ProjectManager
    form_class = modelform_factory(
        model=ProjectManager,
        fields=['start', 'end', 'is_proxy'],
        widgets={'start': CustomDateInput, 'end': CustomDateInput}
    )
    on_success = 'add_manager_success'
    drop_key = 'pk'

    def form_valid(self, form):
        """
        Attempts to save the newly created `ProjectManager` to the database.
        """
        form.instance.manager = User.objects.get(pk=self.kwargs['pk'])
        form.instance.project = Project.objects.get(
            invoice_number=self.kwargs['invoice_number'])
        try:
            form.instance.full_clean()
        except ValidationError as err:
            form.add_error(None, err)
            return self.form_invalid(form)
        return super().form_valid(form)


class AccountingContributorCreateView(AccountingCreateView):
    """
    A class based `CreateView` providing utilities for creating new `ProjectAssignment`s.
    """
    model = ProjectAssignment
    fields = ['salary_level']
    on_success = 'add_contributor_success'
    drop_key = 'pk'

    def form_valid(self, form):
        """
        Attempts to save the newly created `ProjectAssignment` to the database.
        """
        form.instance.contributor = User.objects.get(pk=self.kwargs['pk'])
        form.instance.project = Project.objects.get(
            invoice_number=self.kwargs['invoice_number'])
        try:
            form.instance.full_clean()
        except ValidationError as err:
            form.add_error(None, err)
            return self.form_invalid(form)
        return super().form_valid(form)


class AccountingContributorDeleteView(AccountingDeleteView):
    """
    A class based `DeleteView` providing utilities for deleting `ProjectAssignment`s.
    """
    model = ProjectAssignment
    form_class = ContributorDeleteForm
    on_success = 'delete_contributor_success'
    drop_key = 'pk'

    def get_form_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        kwargs = super().get_form_kwargs()
        kwargs['id'] = self.kwargs['pk']
        return kwargs

class AccountingManagerDeleteView(AccountingDeleteView):
    """
    A class based `DeleteView` providing utilities for deleting `ProjectManager`s.
    """
    model = ProjectManager
    on_success = 'delete_manager_success'
    drop_key = 'pk'


class AccountingWorkhourSheetSelectionView(AccountantRequiredMixin, PeriodSelectorMixin, RedirectView):
    """
    A class based `View` for selecting a `Period` for a work hour sheet.
    """
    def get_redirect_url(self, *args, **kwargs):
        """
        Returns an URL to redirect to based on the given period.
        """
        # Catches Period Selection from GET
        if 'period' in self.request.GET:
            query_set = Period.objects.all()
            form = PeriodSelectForm(query_set, data=self.request.GET)
            if form.is_valid():
                period = form.cleaned_data.get("period")
            else:
                raise Http404()
        # Catches Period no given
        elif 'period_pk' not in kwargs:
            period = Period.objects.latest(create=True)
        # Looks up Period
        else:
            period = get_object_or_404(Period, pk=kwargs['period_pk'])

        return reverse(
            'vkk:workhours:accounting:projects:project:workhours_sheet',
            args=[kwargs['invoice_number'], period.pk]
        )


class AccountingWorkhourSheetView(AccountantRequiredMixin, PeriodSelectorMixin, FormView):
    """
    A class based `View` providing a work hours sheet for a given project and period.
    """
    form_class = WorkhourSheetForm
    template_name = 'vkk/workhours/accounting/projects/project/workhours_sheet.html'
    period_select_namespace = 'vkk:workhours:accounting:projects:project:workhours_sheet_selection'

    def get_form_kwargs(self):
        """
        Returns a dictionary of keyword arguments for instatiating the associated form.
        """
        kwargs = super().get_form_kwargs()
        self.assignments = ProjectAssignment.objects.filter(
            project__invoice_number=self.kwargs['invoice_number']
        ).order_by('contributor__last_name', 'contributor__first_name')
        kwargs.update({
            'period_pk': self.kwargs['period_pk'],
            'assignments': self.assignments,
            'closure_lock': False,
            'invoice_number': self.kwargs['invoice_number'],
        })
        return kwargs

    def get_success_url(self):
        """
        Returns a URL to redirect to after a successful action.
        """
        return reverse(
            'vkk:workhours:accounting:projects:project:workhours_sheet_success',
            args=[
                self.kwargs['invoice_number'],
                self.kwargs['period_pk']
            ]
        )

    def get_context_data(self, **kwargs):
        """
        Returns additional context data as an dictionary. This is used for rendering the page.
        """
        context = super().get_context_data(**kwargs)
        manager_closed_count = self.assignments.filter(
            periodclosure__period=self.kwargs['period_pk'],
            periodclosure__is_closed_manager=True
        ).count()
        assignment_count = self.assignments.count()
        if manager_closed_count == assignment_count:
            context['closed'] = True
        else:
            context['closed'] = False
        context['saveable'] = True

        context['project'] = get_object_or_404(
            Project, invoice_number=self.kwargs['invoice_number'])
        return context

    def form_valid(self, form):
        """
        Saves all changes to work hours to the database.
        """
        form.save()
        return super().form_valid(form)


class AccountingPeriodClosureView(AccountantRequiredMixin, OnSuccessMixin, FormView):
    """
    A class based `View` providing utilities to close the period for new entries 
    for all contributors of the given project.
    """
    template_name = 'vkk/workhours/contributor/closure.html'
    form_class = Form
    on_success = 'period_closure_success'

    def post(self, request, *args, **kwargs):
        """
        Handler for POST rquests.
        """
        form = self.get_form()
        if form.is_valid():
            self.close_period()
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def close_period(self):
        """
        Closes the period for all contributors to a project for a given period.
        """
        assignments = ProjectAssignment.objects.filter(
            project__invoice_number=self.kwargs['invoice_number'],
        )
        period = get_object_or_404(Period, pk=self.kwargs['period_pk'])
        closing = [
            PeriodClosure(
                period=period,
                project_assignment=assignment,
                is_closed_contributor=True,
                is_closed_manager=True
            ) for assignment in assignments
        ]
        PeriodClosure.objects.bulk_create(
            closing,
            update_conflicts=True,
            update_fields=['is_closed_contributor', 'is_closed_manager'],
            unique_fields=['project_assignment', 'period']
        )


class AccountingPeriodClosureSuccessView(AccountantRequiredMixin, CustomSuccessView):
    """
    A class based `View` providing utilities for displaying a successfull closing of a period
    for all contributors to a given project.
    """
    template_name = 'vkk/workhours/contributor/closure_success.html'
    model = PeriodClosure
    on_success = 'workhours_sheet'


# Export

class AccountingExportView(AccountantRequiredMixin, FormView):
    """
    A class based `View` providing an overview over all possible actions associated with the
    export of project related data.
    """
    template_name = 'vkk/workhours/accounting/projects/project/export.html'
