"""
A collection of `Form`s for this module.
"""

from django import forms
from django.db.models import Sum
from django.utils import timezone
from django.utils.formats import date_format
from vkk.users.models import User
from vkk.workhours.models import ProjectAssignment, Period, WorkHours, WorkHoursCorrection, PeriodClosure
from django.utils.translation import gettext_lazy as _


class YearSelectForm(forms.Form):
    """
    A `Form` subclass for selecting a year.
    """

    def __init__(self, *args, **kwargs):
        """
        Constructs and returns an instance of this class.
        """
        super().__init__(*args, **kwargs)
        choices = {(entry['start'].year, entry['start'].year)
                   for entry in Period.objects.all().values('start')}
        choices = list(choices)
        choices.sort()
        self.fields['year'] = forms.ChoiceField(
            label=_('Year'),
            choices=choices
        )


class AssigneeForm(forms.Form):
    """
    A `Form` subclass providing functionality for administrating work hours of a contributor.
    """
    custom_template = 'vkk/workhours/allprojects/assignee_form.html'

    class Media:
        """
        A inner class for meta information.
        """
        css = {'all': ('styles/assignee_form.css',)}
        js = ('scripts/assignee.js',)

    def __init__(self, *args, assignee=None, manager=None, year=None, **kwargs):
        """
        Constructs and returns an instance of this class.
        """
        super().__init__(*args, **kwargs)
        self._assignee = assignee
        self._year = year
        self._periods = Period.objects.filter(
            start__year=year,
        ).order_by('start')
        self._assignments = ProjectAssignment.objects.filter(
            contributor=assignee,
            project__projectmanager__manager=manager,
            project__projectmanager__end__gte=timezone.now().date()
        ).select_related("project").order_by('project__invoice_number')

        for assignment in self._assignments:
            for period in self._periods:
                workhours = WorkHours.objects.filter(
                    project_assignment=assignment,
                    period=period,
                )
                workhours_correction = WorkHoursCorrection.objects.filter(
                    project_assignment=assignment,
                    period=period,
                )

                disabled = timezone.now() > period.dead_line_final
                if not disabled:
                    closure = PeriodClosure.objects.filter(
                        project_assignment=assignment,
                        period=period
                    )
                    if closure.exists():
                        disabled = closure[0].is_closed_manager

                if not disabled:
                    disabled = period.end < assignment.project.start
                    disabled |= period.start > assignment.project.end

                month = date_format(period.start, format='F', use_l10n=True)
                label = f"{assignment.project.name} {month}"

                field = forms.DecimalField(
                    required=False,
                    min_value=0,
                    max_digits=5,
                    decimal_places=2,
                    disabled=disabled,
                    label=label
                )

                initial = workhours.aggregate(Sum('hours'))['hours__sum'] or 0
                if workhours_correction.exists():
                    initial += workhours_correction[0].ammount
                if initial != 0:
                    field.initial = initial

                self.fields[str(assignment.id) + '_' + str(period.id)] = field

    def _get_field_structure(self):
        return [
            (
                assignment,
                [
                    self[str(assignment.pk) + '_' + str(period.id)] for period in self._periods
                ]
            ) for assignment in self._assignments
        ]

    def as_html(self):
        context = super().get_context()
        context.update({
            'assignee': self._assignee,
            'year': self._year,
            'periods': self._periods,
            'fields_more': self._get_field_structure()
        })
        return self.render(
            template_name=self.custom_template,
            context=context
        )

    def save(self):
        if self.is_valid() and self.has_changed():
            for assignment in self._assignments:
                for period in self._periods:
                    field_name = str(assignment.pk) + '_' + str(period.id)
                    if field_name in self.changed_data:
                        workhours = WorkHours.objects.filter(
                            project_assignment=assignment,
                            period=period,
                        )
                        value = (self.cleaned_data[field_name] or 0) - (
                            workhours.aggregate(Sum('hours'))['hours__sum'] or 0)
                        if value == 0:
                            WorkHoursCorrection.objects.filter(
                                project_assignment=assignment,
                                period=period,
                            ).delete()
                        else:
                            WorkHoursCorrection.objects.update_or_create(
                                project_assignment=assignment,
                                period=period,
                                defaults={'ammount': value}
                            )


class AssgineeClosureForm(forms.Form):
    assignee = forms.ModelChoiceField(queryset=User.objects.all())
    period = forms.ModelChoiceField(queryset=Period.objects.all())
    manager = forms.ModelChoiceField(queryset=User.objects.all())

    def save(self):
        if self.is_valid():
            assignee = self.cleaned_data['assignee']
            period = self.cleaned_data['period']
            manager = self.cleaned_data['manager']
            project_assignments = ProjectAssignment.objects.filter(
                contributor=assignee,
                project__projectmanager__manager=manager
            )

            objs = [PeriodClosure(
                period=period,
                project_assignment=project_assignment,
                is_closed_contributor=True,
                is_closed_manager=True,
            ) for project_assignment in project_assignments]

            PeriodClosure.objects.bulk_create(
                objs=objs,
                ignore_conflicts=False,
                update_conflicts=True,
                update_fields=['is_closed_contributor', 'is_closed_manager'],
                unique_fields=['period', 'project_assignment']
            )
