"""
This submodule contains all of the class based views of this module.
"""

from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView, RedirectView
from django.urls import reverse
from django.db.models import Max
from django.utils import timezone
from vkk.generic.views import CustomListView, CustomSuccessView
from vkk.users.models import User
from vkk.workhours.models import Period
from .mixins import ProjectManagerRequiredMixin
from .forms import AssigneeForm, YearSelectForm, AssgineeClosureForm


class AllProjectYearRedirectView(ProjectManagerRequiredMixin, RedirectView):
    """
    A class based `View` providing a functionality for selecting and redirecting for a given year.
    """
    def get_redirect_url(self, *args, **kwargs):
        """
        Returns an URL for redirecting.
        """
        # Catches Period Selection from GET
        if 'year' in self.request.GET:
            form = YearSelectForm(data=self.request.GET)
            if form.is_valid():
                year = int(form.cleaned_data['year'])
        else:
            year = Period.objects.aggregate(Max('start'))['start__max'].year
        return reverse('vkk:workhours:allprojects:overview', kwargs={'year': year})


class AllProjectAssigneesOverView(ProjectManagerRequiredMixin, CustomListView):
    """
    A class based `View` providing functionality for listing all contributors to be 
    managed by the given project manager no matter the specific project.
    """
    model = User
    fields = ['first_name', 'last_name', 'email']
    action_options = {'update': _('Details')}
    ordering = ['last_name', 'first_name']
    template_name = 'vkk/workhours/allprojects/overview.html'

    def get_context_data(self, **kwargs):
        """
        Returns a context for rendering the page.
        """
        context = super().get_context_data(**kwargs)
        context["form"] = YearSelectForm(
            initial={'year': self.kwargs['year']}
        )
        return context

    def get_queryset(self):
        """
        Returns a query set containing all contributors which are assigned to projects, which
        are managed by the given user as a project manager.
        """
        query_set = super().get_queryset().filter(
            projectassignment__project__projectmanager__manager=self.request.user,
            projectassignment__project__projectmanager__end__gte=timezone.now().date(),
        ).distinct()
        return query_set


class AssigneeUpdate(ProjectManagerRequiredMixin, FormView):
    """
    A class based `View` providing functionality for updating work hours 
    of a selected contributor.
    """
    form_class = AssigneeForm
    template_name = 'vkk/workhours/allprojects/assignee_update.html'

    def get_form_kwargs(self):
        """
        Returns the keyword arguments meant for initializing the associated `Form`.
        """
        kwargs = super().get_form_kwargs()
        kwargs.update({
            'assignee': self.kwargs['pk'],
            'manager': self.request.user,
            'year': self.kwargs['year'],
        })
        return kwargs

    def get_context_data(self, **kwargs):
        """
        Returns a context for rendering the page.
        """
        context = super().get_context_data(**kwargs)
        context["assignee"] = User.objects.get(id=self.kwargs['pk'])
        return context

    def get_success_url(self):
        """
        Returns an URL to redirect to after a successfull action.
        """
        return reverse(
            'vkk:workhours:allprojects:update_success', kwargs={'year': self.kwargs['year']},
        )

    def form_valid(self, form):
        """
        Saves the Information associated with the `Form`, after valididating it, and returns
        said instance. 
        """
        form.save()
        return super().form_valid(form)


class AssigneeClosureView(ProjectManagerRequiredMixin, FormView):
    """
    A class based `View` providing functionality for closing a period for a contributor.
    """
    form_class = AssgineeClosureForm
    template_name = 'vkk/workhours/contributor/closure.html'

    def post(self, request, *args, **kwargs):
        """
        Handler for POST request. Returns either a 403 response or redirection after
        a sucsessful action.
        """
        form = self.form_class(
            data={
                'period': self.kwargs['period_pk'],
                'assignee': self.kwargs['assignee_pk'],
                'manager': self.request.user
            },
        )
        if form.is_valid:
            form.save()
            return self.form_valid(form)
        else:
            return self.handle_no_permission()

    def get_context_data(self, **kwargs):
        """
        Returns a context for rednering the page.
        """
        context = super().get_context_data(**kwargs)
        context.pop('form')
        return context

    def get_success_url(self):
        """
        Returns an URL to redirect to after a successful action.
        """
        url = reverse(
            'vkk:workhours:allprojects:closure_success',
            args=[self.kwargs['year'], self.kwargs['assignee_pk']],
        )
        return url


class AssigneeUpdateSuccessView(ProjectManagerRequiredMixin, CustomSuccessView):
    """
    A class based `View` providing a success page after updating a contributors workhours.
    """
    pass


class AssigneeClosureSuccessView(ProjectManagerRequiredMixin, CustomSuccessView):
    """
    A class based `View` providing a success page after closing a contributors period.
    """
    def get_success_url(self):
        """
        Returns a URL for a hyperlink.
        """
        url = reverse(
            'vkk:workhours:allprojects:update',
            args=[self.kwargs['year'], self.kwargs['assignee_pk']],
        )
        return url
