"""
A collection of `Form`s for contributors.
"""

import datetime
from django import forms
from vkk.workhours.models import WorkHours, Period, ProjectAssignment
from vkk.workhours.forms import date_iterator


class WorkhourCalendarForm(forms.Form):
    """
    A form for entering work hours resembling a calender.
    """
    class Media:
        """
        An inner class containing meta information for styling.
        """
        css = {'all': ('styles/calendar.css',)}

    template_name_calendar = 'vkk/workhours/contributor/calendar.html'

    def __init__(self, *args, period_pk=None, assignment_pk=None, **kwargs):
        """
        Constructs an `WorkhourCalendarForm` instance using the given primary
         keys of an `Period` and `ProjectAssignment`.
        """
        super().__init__(*args, **kwargs)
        self._project_assignment = ProjectAssignment.objects.get(
            pk=assignment_pk
        )

        # create fields
        self._period = Period.objects.get(pk=period_pk)
        today = datetime.date.today()
        field_class = WorkHours.hours.field.formfield

        for date in date_iterator(self._period.start, self._period.end):
            is_disabled = date > today or self._project_assignment.project.start > date \
                or self._project_assignment.project.end < date
            field = field_class(
                min_value=0,
                max_value=24,
                required=False,
                label=date.day,
                label_suffix='',
                disabled=is_disabled
            )
            self.fields[date.isoformat()] = field

        # provide initial data
        workhours = WorkHours.objects.filter(
            period__pk=period_pk,
            project_assignment__pk=assignment_pk
        )
        for entry in workhours:
            field = self.fields.get(entry.day.isoformat())
            if field is not None:
                field.initial = entry.hours

    def _get_date_calendar_structure(self):
        """
        Private method used for structuring some data for rendering the form.
        """
        dates = date_iterator(self._period.start, self._period.end)

        # subdevision into months
        months = []
        month_number = -1
        for date in dates:
            if date.month != month_number:
                months += [[]]
                month_number = date.month
            months[-1] += [date]

        # subdevision into weeks
        weeks = []
        for month in months:
            weeks += [[]]
            week_day = 7
            for date in month:
                if date.weekday() < week_day:
                    weeks[-1] += [[]]
                weeks[-1][-1] += [date]
                week_day = date.weekday()

        return weeks

    def _get_field_calendar_structure(self):
        """
        Private method used for structuring some data for rendering the form.
        """
        calendar = self._get_date_calendar_structure()
        return [[(
            [(day, self[day.isoformat()]) for day in week]
        ) for week in month] for month in calendar]

    def as_calendar(self):
        """
        Renders the form resembling a calender.
        """
        context = super().get_context()
        context.update({'fields_more': self._get_field_calendar_structure()})
        return self.render(
            template_name=self.template_name_calendar,
            context=context
        )

    def save(self):
        """
        Tries to create, modify or delete all changes.
        """
        if self.is_valid() and self.has_changed():
            add = []
            delete = []
            for field_name in self.changed_data:
                initial = self.fields[field_name].initial
                value = self.cleaned_data.get(field_name)
                if value is None or value == 0.0:
                    delete.append(
                        datetime.date.fromisoformat(field_name)
                    )
                elif initial is None:
                    add.append(WorkHours(
                        project_assignment=self._project_assignment,
                        period=self._period,
                        day=datetime.date.fromisoformat(field_name),
                        hours=value
                    ))
                else:
                    WorkHours.objects.filter(
                        project_assignment=self._project_assignment,
                        period=self._period,
                        day=datetime.date.fromisoformat(field_name),
                    ).update(hours=value)
            WorkHours.objects.filter(
                project_assignment=self._project_assignment,
                period=self._period,
                day__in=delete
            ).delete()
            WorkHours.objects.bulk_create(add)
