"""
A collection of forms used throughout the `workhours` module.
"""

import datetime
from django import forms
from django.db.models import Sum
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from vkk.workhours.models import WorkHours, WorkHoursCorrection, Period, PeriodClosure, Project


def date_iterator(start, end):
    """
    Returns an iterator over a range of dates.
    """
    delta = end - start
    for day in range(delta.days+1):
        yield start + datetime.timedelta(day)


class CustomDateInput(forms.DateInput):
    """
    This subclass of `DateInput` provides the HTML input type "date" for widgets.
    """
    input_type = 'date'

    def format_value(self, value):
        """
        Returns a date in ISO-8601 format.
        """
        if isinstance(value, datetime.date):
            return value.isoformat()
        else:
            return super().format_value(value)


class CustomDateTimeInput(forms.DateTimeInput):
    """
    This subclass of `DateTimeInput` provides the HTML input type "datetime-local" 
    for widgets.
    """
    input_type = 'datetime-local'


class PeriodSelectForm(forms.Form):
    """
    A form for selecting a `Period` instance.
    """

    def __init__(self, queryset, *args, **kwargs):
        """
        Constructs a `PeriodSelectForm` object.
        """
        super().__init__(*args, **kwargs)
        self.fields["period"] = forms.ModelChoiceField(
            queryset,
            empty_label=None,
            label=_('Period')
        )


class WorkhourSheetForm(forms.Form):
    """
    A form closely resembling a sheet of paper for keeping track of hour worked
    on a project by individual contributors.
    """
    template_name_sheet = 'vkk/workhours/workhours_sheet_form.html'
    sums = dict()
    closures = dict()

    def __init__(self, *args, period_pk=None, assignments=None, closure_lock=True, invoice_number=None, **kwargs):
        """
        This constructor takes a primary key value of a `Period` instance, an
         `ProjectAssingment` instance, a projects invoice number and whether
         the inputs should be disabled according to `PeriodClosure`instances.

        Returns a newly constructed `WorkhourSheetForm` object.
        """
        super().__init__(*args, **kwargs)

        today = datetime.date.today()
        field_class = WorkHours.hours.field.formfield
        field_class_correction = WorkHoursCorrection.ammount.field.formfield

        self._period = Period.objects.get(pk=period_pk)
        self._dates = list(date_iterator(self._period.start, self._period.end))
        self._assignments = assignments.select_related('contributor')
        self._closure = PeriodClosure.objects.filter(
            period=self._period,
        )
        self._project = None
        if invoice_number is not None:
            projects = Project.objects.filter(invoice_number=invoice_number)
            if projects.exists():
                self._project = projects[0]

        for assignment in self._assignments:
            closed = closure_lock and self._period.dead_line < timezone.now()
            closure = self._closure.filter(
                project_assignment=assignment
            )
            if closure_lock and not closed and closure.exists():
                closed = closure[0].is_closed_manager
            expired = False
            for date in self._dates:
                if self._project is not None:
                    expired = self._project.start > date or self._project.end < date
                # create fields for each assignment
                is_future = date > today
                field = field_class(
                    min_value=0,
                    max_value=24,
                    required=False,
                    label=date.day,
                    label_suffix='',
                    disabled=(is_future or closed or expired)
                )
                self.fields[str(assignment.pk) + '_' +
                            date.isoformat()] = field
            self.fields[str(assignment.pk) + '_C'] = field = field_class_correction(
                label=_('C'),
                required=False,
                label_suffix='',
                disabled=closed,
            )

            # provide initial data
            workhours = WorkHours.objects.filter(
                period__pk=period_pk,
                project_assignment=assignment
            )
            for entry in workhours:
                field = self.fields.get(
                    str(assignment.pk) + '_' + entry.day.isoformat())
                if field is not None:
                    field.initial = entry.hours

            assignment_sum = workhours.aggregate(Sum('hours'))
            self.sums[str(assignment.pk)] = assignment_sum['hours__sum'] or 0

            correction = WorkHoursCorrection.objects.filter(
                period__pk=period_pk,
                project_assignment=assignment
            )
            if correction.exists():
                field = self.fields.get(str(assignment.pk) + '_C')
                if field is not None:
                    field.initial = correction[0].ammount
                self.sums[str(assignment.pk)] += correction[0].ammount

            # feedback for closed periods
            feedback = ""
            if closure.exists() and closure[0].is_closed_contributor:
                feedback += "☺️"
            if closure.exists() and closure[0].is_closed_manager:
                feedback += "☺️"
            self.closures[str(assignment.pk) + '_closures'] = feedback

    def _get_field_sheet_structure(self):
        """
        Returns a list for building the HTML representation of the form. 
        """
        return [
            (
                assignment,
                [
                    self[str(assignment.pk) + '_' + date.isoformat()] for date in self._dates
                ] + [self[str(assignment.pk) + '_C'], self.sums[str(assignment.pk)], self.closures[str(assignment.pk) + '_closures']]
            ) for assignment in self._assignments
        ]

    def as_sheet(self):
        """
        Renders and returns the HTML form.
        """
        context = super().get_context()
        context.update({
            'start': self._period.start,
            'fields_more': self._get_field_sheet_structure()
        })
        return self.render(
            template_name=self.template_name_sheet,
            context=context
        )

    def save(self):
        """
        Modifies and saves all associated `WorkHours` instances.
        """
        if self.is_valid() and self.has_changed():
            for assignment in self._assignments:
                add = []
                delete = []
                field_names = [(date, str(assignment.pk) + '_' +
                                date.isoformat()) for date in self._dates]
                for date, field_name in field_names:
                    if field_name in self.changed_data:
                        initial = self.fields[field_name].initial
                        value = self.cleaned_data.get(field_name)
                        if value is None or value == 0.0:
                            delete.append(
                                date
                            )
                        elif initial is None:
                            add.append(WorkHours(
                                project_assignment=assignment,
                                period=self._period,
                                day=date,
                                hours=value
                            ))
                        else:
                            WorkHours.objects.filter(
                                project_assignment=assignment,
                                period=self._period,
                                day=date,
                            ).update(hours=value)
                WorkHours.objects.filter(
                    project_assignment=assignment,
                    period=self._period,
                    day__in=delete
                ).delete()
                WorkHours.objects.bulk_create(add)
                correction_name = str(assignment.pk) + '_C'
                if correction_name in self.changed_data:
                    value = self.cleaned_data.get(correction_name)
                    if value is None or value == 0.0:
                        WorkHoursCorrection.objects.filter(
                            project_assignment=assignment,
                            period=self._period
                        ).delete()
                    else:
                        WorkHoursCorrection.objects.update_or_create(
                            project_assignment=assignment,
                            period=self._period,
                            defaults={'ammount': value}
                        )
