"""
This submodule contains all of the class based views of this module.
"""

from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError
from django.urls import reverse
from django.forms import Form
from django.shortcuts import get_object_or_404
from django.views.generic import FormView, RedirectView
from vkk.generic.views import CustomDetailView, CustomUpdateView, CustomSuccessView, CustomCreateView, CustomFilterView
from vkk.workhours.models import Project, ProjectManager, ProjectAssignment, Period, PeriodClosure
from vkk.workhours.mixins import PeriodSelectorMixin
from vkk.generic.mixins import OnSuccessMixin
from vkk.workhours.forms import WorkhourSheetForm, PeriodSelectForm
from vkk.users.models import User, send_registration_mail
from .mixins import ManagerRequiredMixin
from .forms import ProjectRegisterContributorForm


class ManagerProjectOverView(ManagerRequiredMixin, CustomDetailView):
    """
    A class based view, which provides an overview of a project to a project manager.
    """
    model = Project
    slug_field = 'invoice_number'
    slug_url_kwarg = 'invoice_number'
    template_name = 'vkk/workhours/manager/details.html'
    fields = ['name', 'invoice_number',
              'department', 'contractor', 'start', 'end']

    def get_context_data(self, **kwargs):
        """
        Returns a dictionary of data meant for the template layer.
        """
        context = super().get_context_data(**kwargs)
        context["managers"] = ProjectManager.objects.filter(
            project__invoice_number=self.kwargs['invoice_number'],
            manager__is_active=True,
        ).prefetch_related('manager').order_by('manager__last_name', 'manager__first_name')
        context["contributors"] = ProjectAssignment.objects.filter(
            project__invoice_number=self.kwargs['invoice_number'],
            contributor__is_active=True,
        ).prefetch_related('contributor').order_by('contributor__last_name', 'contributor__first_name')
        return context


class ManagerDetailView(ManagerRequiredMixin, CustomDetailView):
    """
    A class based view extending the `CustomDetailView` class with checks for user permissions.
    """
    pass


class ManagerUpdateView(ManagerRequiredMixin, CustomUpdateView):
    """
    A class based view extending the `CustomUpdateView` class with checks for user permissions.
    """
    pass


class ManagerSuccessView(ManagerRequiredMixin, CustomSuccessView):
    """
    A class based view extending the `CustomSuccessView` class with checks for user permissions.
    """
    pass


class ManagerRegisterContributorView(ManagerRequiredMixin, CustomCreateView):
    """
    A class based view providing functionality for registering new users.
    """
    model = User
    form_class = ProjectRegisterContributorForm
    on_success = 'register_contributor_success'

    def form_valid(self, form):
        """
        After the associated form has been validated, a new `User` is created and
         assigned to the project. Returns a `HTTPResponse` object.
        """
        user = form.save()
        send_registration_mail(user)
        project = Project.objects.get(
            invoice_number=self.kwargs['invoice_number'])
        ProjectAssignment.objects.create(
            contributor=user,
            project=project,
            salary_level=form.cleaned_data['salary_level']
        )
        return super().form_valid(form)


class ManagerContributorCreateView(ManagerRequiredMixin, CustomCreateView):
    """
    A class based view providing functionality for assigning users to the project.
    """
    model = ProjectAssignment
    fields = ['salary_level']
    on_success = 'add_contributor_success'
    drop_key = 'pk'

    def form_valid(self, form):
        """
        After the associated form has been validated, the given user is
         assigned to the project. Returns a `HTTPResponse` object.
        """
        form.instance.contributor = User.objects.get(pk=self.kwargs['pk'])
        form.instance.project = Project.objects.get(
            invoice_number=self.kwargs['invoice_number'])
        try:
            form.instance.full_clean()
        except ValidationError as err:
            form.add_error(None, err)
            return self.form_invalid(form)
        return super().form_valid(form)


class ManagerContributorFilterView(ManagerRequiredMixin, CustomFilterView):
    """
    Extends the `CustomFilterView` to offer a view for listing and filtering potential new contributors to the given project.
    """
    model = User
    fields = ['last_name', 'first_name', 'email']
    action_options = {'add_contributor': _('Add')}

    def get_queryset(self):
        """
        Returns the query set of all `Users` excluding those, who are already contributors to the given project.
        """
        return super().get_queryset().exclude(
            projectassignment__project__invoice_number=self.kwargs['invoice_number']
        )


class ManagerWorkhourSheetSelectionView(ManagerRequiredMixin, PeriodSelectorMixin, RedirectView):
    """
    A class based view providing functionality for selecting a and redirecting to specific work hours sheet.
    """

    def get_redirect_url(self, *args, **kwargs):
        """
        Returns an URL to redirect to based of whether a valid `Period` is provided.
        """
        # Catches Period Selection from GET
        if 'period' in self.request.GET:
            query_set = Period.objects.all()
            form = PeriodSelectForm(query_set, data=self.request.GET)
            if form.is_valid():
                period = form.cleaned_data.get("period")
            else:
                raise Http404()
        # Catches Period no given
        elif 'period_pk' not in kwargs:
            period = Period.objects.latest(create=True)
        # Looks up Period
        else:
            period = get_object_or_404(Period, pk=kwargs['period_pk'])

        return reverse(
            'vkk:workhours:manager:workhours_sheet',
            args=[kwargs['invoice_number'], period.pk]
        )


class ManagerWorkhourSheetView(ManagerRequiredMixin, PeriodSelectorMixin, FormView):
    """
    A class based view providing functionality for displaying and managing a work hours sheet at a form.
    """
    form_class = WorkhourSheetForm
    template_name = 'vkk/workhours/workhours_sheet.html'
    period_select_namespace = 'vkk:workhours:manager:workhours_sheet_selection'

    def setup(self, request, *args, **kwargs):
        """
        Extends the `setup()` method of the parent class.
        """
        super().setup(request, *args, **kwargs)
        self.assignments = ProjectAssignment.objects.filter(
            project__invoice_number=self.kwargs['invoice_number']
        )

    def get_context_data(self, **kwargs):
        """
        Returns a dictionary with context data for the template layer.
        """
        context = super().get_context_data(**kwargs)
        manager_closed_count = self.assignments.filter(
            periodclosure__period=self.kwargs['period_pk'],
            periodclosure__is_closed_manager=True
        ).count()
        assignment_count = self.assignments.count()
        if manager_closed_count == assignment_count:
            context['closed'] = True
            context['saveable'] = False
        else:
            context['closed'] = False
            context['saveable'] = True

        context['project'] = get_object_or_404(
            Project, invoice_number=self.kwargs['invoice_number'])
        return context

    def get_form_kwargs(self):
        """
        Returns a dictionary with keyword arguments for instantiating the form class.
        """
        kwargs = super().get_form_kwargs()
        assignments = self.assignments.order_by(
            'contributor__last_name', 'contributor__first_name')
        kwargs.update({
            'period_pk': self.kwargs['period_pk'],
            'assignments': assignments,
            'invoice_number': self.kwargs['invoice_number']
        })
        return kwargs

    def get_success_url(self):
        """
        Returns a URL to redirect to after accepting and processing the form.
        """
        return reverse(
            'vkk:workhours:manager:workhours_sheet_success',
            args=[
                self.kwargs['invoice_number'],
                self.kwargs['period_pk']
            ]
        )

    def form_valid(self, form):
        """
        Calls the `save()` method on the associated form and returns a redirect response.
        """
        form.save()
        return super().form_valid(form)


class ManagerPeriodClosureView(ManagerRequiredMixin, OnSuccessMixin, FormView):
    """
    A class based view providing functionality to lock work hour entries of all
     `ProjectAssignment`s associated with the given project for a given `Period`.
    """
    template_name = 'vkk/workhours/contributor/closure.html'
    form_class = Form
    on_success = 'period_closure_success'

    def post(self, request, *args, **kwargs):
        """
        Extends the `post()` method with functionality associated with the form handling.
        """
        form = self.get_form()
        if form.is_valid():
            self.close_period()
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def close_period(self):
        """
        Manages `PeriodClosure` instances accordingly.
        """
        assignments = ProjectAssignment.objects.filter(
            project__invoice_number=self.kwargs['invoice_number'],
        )
        period = get_object_or_404(Period, pk=self.kwargs['period_pk'])
        closing = [
            PeriodClosure(
                period=period,
                project_assignment=assignment,
                is_closed_contributor=True,
                is_closed_manager=True
            ) for assignment in assignments
        ]
        PeriodClosure.objects.bulk_create(
            closing,
            update_conflicts=True,
            update_fields=['is_closed_contributor', 'is_closed_manager'],
            unique_fields=['project_assignment', 'period']
        )


class ManagerPeriodClosureSuccessView(ManagerRequiredMixin, CustomSuccessView):
    """
    Extends the `CustomSuccessView` class to provide functionality for
     displaying a successful lock of a work hours sheet.
    """
    template_name = 'vkk/workhours/contributor/closure_success.html'
    model = PeriodClosure
    on_success = 'workhours_sheet'
